from collections import defaultdict
from typing import Dict, Type

from alpaca.common import HTTPResult, RawData

"""
These functions were created and put in this file to handle all of the edge cases
for parsing data that exist in the market data API. 

v2/stocks and v1beta2/crypto 
"""


def parse_obj_as_symbol_dict(model: Type, raw_data: RawData) -> Dict[str, Type]:
    """
    Parses raw_data into a dictionary where the keys are the string valued symbols and the values are the
    parsed data into the model.

    Args:
        model (Type): The model we want to parse the data into
        raw_data (RawData): The raw data from the API

    Returns:
        Dict[str, Type]: The symbol keyed dictionary of parsed data
    """
    if raw_data is None:
        return {}
    return {
        k: model(symbol=k, raw_data=v) for k, v in raw_data.items() if v is not None
    }


def get_data_from_response(response: HTTPResult) -> RawData:
    """
    From the raw API response, retrieves the field that contains market data.

    Args:
        response (HTTPResult): The raw API response

    Returns:
        RawData: The market data from the response
    """
    data_keys = {
        "trade",
        "trades",
        "quote",
        "quotes",
        "bar",
        "bars",
        "snapshot",
        "snapshots",
        "orderbook",
        "orderbooks",
    }

    selected_key = data_keys.intersection(response)

    if selected_key is None or len(selected_key) < 1:
        raise ValueError("The data in response does not match any known keys.")

    # assume selected_key only contains 1 value
    selected_key = selected_key.pop()

    # formatting a single symbol response so that this method
    # always returns a symbol keyed data dictionary
    if "symbol" in response:
        return {response["symbol"]: response[selected_key]}

    return response[selected_key]


def format_dataset_response(
    response: HTTPResult, data_by_symbol: defaultdict
) -> defaultdict:
    """
    Formats data from BarSet, TradeSet endpoints. Uses defaultdict for simpler syntax when
    doing list comprehension.

    Args:
        response (HTTPResult): The raw response from bars, trades, quotes, endpoint
        data_by_symbol (defaultdict):  The dictionary we want to format the raw response into

    Returns:
        defaultdict: The dictionary populated with data
    """
    # data_by_symbol is in format of
    #    {
    #       "symbol1": [ "data1", "data2", ... ],
    #       "symbol2": [ "data1", "data2", ... ],
    #                ....
    #    }

    response_data = get_data_from_response(response)

    # add elements to data_by_symbol
    # for list data types just extend
    # for non-list types, add as element of a list.
    # list comprehension used for speed
    [
        (
            data_by_symbol[symbol].extend(data)
            if isinstance(data, list)
            else data_by_symbol[symbol].append(data)
        )
        for symbol, data in response_data.items()
    ]

    return data_by_symbol


def format_latest_data_response(
    response: HTTPResult, data_by_symbol: defaultdict
) -> defaultdict:
    """
    Parses data from the "latest" endpoints and populates the data_by_symbol dict. (latest_quote, latest_bar). Also includes crypto/snapshots endpoint,
    but not stocks/snapshots.

    Args:
        response (HTTPResult): The response from the latest data endpoint.
        data_by_symbol (defaultdict): The dictionary we want to format the raw response into

    Returns:
        defaultdict: The dictionary populated with latest data points
    """
    response_data = get_data_from_response(response)

    for symbol, data in response_data.items():
        data_by_symbol[symbol] = data

    return data_by_symbol


def format_snapshot_data(
    response: HTTPResult, data_by_symbol: defaultdict
) -> defaultdict:
    """
    Formats data from stocks/snapshot endpoint. Exists because v2/stocks/snapshot
    is different from v1beta2/crypto/snapshot.

    Args:
        response (HTTPResult): The raw response from v2/stocks/snapshot endpoint
        data_by_symbol (defaultdict):  The dictionary we want to format the raw response into

    Returns:
        defaultdict: The dictionary populated with data
    """
    # TODO: Improve snapshot parsing.
    if "symbol" in response:
        symbol = response["symbol"]
        del response["symbol"]
        data_by_symbol[symbol] = response
    else:
        if "snapshots" in response:
            response = response["snapshots"]
        for symbol, data in response.items():
            data_by_symbol[symbol] = data

    return data_by_symbol



from alpaca.data.historical.crypto import CryptoHistoricalDataClient
from alpaca.data.historical.news import NewsClient
from alpaca.data.historical.option import OptionHistoricalDataClient
from alpaca.data.historical.screener import ScreenerClient
from alpaca.data.historical.stock import StockHistoricalDataClient

__all__ = [
    "CryptoHistoricalDataClient",
    "StockHistoricalDataClient",
    "NewsClient",
    "OptionHistoricalDataClient",
    "ScreenerClient",
]



from datetime import datetime
from typing import List

from alpaca.common.models import ValidateBaseModel as BaseModel
from alpaca.data.enums import MarketType


class ActiveStock(BaseModel):
    """
    Represent one asset that was a most active on the most actives endpoint.

    Attributes:
        symbol (str): Symbol of market moving asset.
        volume (float): Cumulative volume for the current trading day.
        trade_count (float): Cumulative trade count for the current trading day.
    """

    symbol: str
    volume: float
    trade_count: float


class MostActives(BaseModel):
    """
    Represent the response model for the MostActives endpoint.
    Attributes:
        most_actives (List[ActiveStock]): list of top N most active symbols.
        last_updated (datetime):
            Time when the MostActives were last computed.
            Formatted as a RFC 3339 formatted datetime with nanosecond precision.
    """

    most_actives: List[ActiveStock]
    last_updated: datetime


class Mover(BaseModel):
    """
    Represent one asset that was a top mover on the top market movers endpoint.
    Attributes:
        symbol (str): Symbol of market moving asset.
        percent_change (float): Percentage difference change for the day.
        change (float): Difference in change for the day.
        price (float): Current price of market moving asset.
    """

    symbol: str
    percent_change: float
    change: float
    price: float


class Movers(BaseModel):
    """
    Represent the response model for the top market movers endpoint.
    Attributes:
        gainers (List[Mover]): list of top N gainers.
        losers (List[Mover]): list of top N losers.
        market_type (MarketType): Market type (stocks or crypto).
        last_updated (datetime):
            Time when the movers were last computed.
            Formatted as a RFC 3339 formatted datetime with nanosecond precision.
    """

    gainers: List[Mover]
    losers: List[Mover]
    market_type: MarketType
    last_updated: datetime



from typing import Dict

BAR_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "o": "open",
    "h": "high",
    "l": "low",
    "c": "close",
    "v": "volume",
    "n": "trade_count",
    "vw": "vwap",
}

QUOTE_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "ax": "ask_exchange",
    "ap": "ask_price",
    "as": "ask_size",
    "bx": "bid_exchange",
    "bp": "bid_price",
    "bs": "bid_size",
    "c": "conditions",
    "z": "tape",
}

TRADE_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "p": "price",
    "s": "size",
    "x": "exchange",
    "i": "id",
    "c": "conditions",
    "z": "tape",
}

SNAPSHOT_MAPPING: Dict[str, str] = {
    "latestTrade": "latest_trade",
    "latestQuote": "latest_quote",
    "minuteBar": "minute_bar",
    "dailyBar": "daily_bar",
    "prevDailyBar": "previous_daily_bar",
    "impliedVolatility": "implied_volatility",
    "greeks": "greeks",
}

ORDERBOOK_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "b": "bids",
    "a": "asks",
    "r": "reset",
}

TRADING_STATUS_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "sc": "status_code",
    "sm": "status_message",
    "rc": "reason_code",
    "rm": "reason_message",
    "z": "tape",
}

TRADE_CANCEL_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "p": "price",
    "s": "size",
    "x": "exchange",
    "i": "id",
    "a": "action",
    "z": "tape",
}

TRADE_CORRECTION_MAPPING: Dict[str, str] = {
    "t": "timestamp",
    "x": "exchange",
    "oi": "original_id",
    "op": "original_price",
    "os": "original_size",
    "oc": "original_conditions",
    "ci": "corrected_id",
    "cp": "corrected_price",
    "cs": "corrected_size",
    "cc": "corrected_conditions",
    "z": "tape",
}



from typing import Optional
from alpaca.trading.models import Order as BaseOrder


class Order(BaseOrder):
    """
    See base alpaca.trading.models.Order model for full list of base attributes.

    Attributes:
        commission (float): The dollar value commission you want to charge the end user.
    """

    commission: Optional[float] = None



from enum import Enum


class Exchange(str, Enum):
    """The exchanges that provide data feeds to Alpaca.

    Attributes:
        Z (str): Cboe BZ
        I (str): International Securities Exchange
        M (str): Chicago Stock Exchange
        U (str): Members Exchange
        L (str): Long Term Stock Exchange
        W (str): CBOE
        X (str): NASDAQ OMX PSX
        B (str): NASDAQ OMX BX
        D (str): FINRA ADF
        J (str): Cboe EDGA
        P (str): NYSE Arca
        Q (str): NASDAQ OMX
        S (str): NASDAQ Small Cap
        V (str): IEX
        A (str): NYSE American (AMEX)
        E (str): Market Independent
        N (str): New York Stock Exchange
        T (str): NASDAQ Int
        Y (str): Cboe BYX
        C (str): National Stock Exchange
        H (str): MIAX
        K (str): Cboe EDGX
    """

    Z = "Z"
    I = "I"
    M = "M"
    U = "U"
    L = "L"
    W = "W"
    X = "X"
    B = "B"
    D = "D"
    J = "J"
    P = "P"
    Q = "Q"
    S = "S"
    V = "V"
    A = "A"
    E = "E"
    N = "N"
    T = "T"
    Y = "Y"
    C = "C"
    H = "H"
    K = "K"


class DataFeed(str, Enum):
    """Equity market data feeds. OTC and SIP are available with premium data subscriptions.

    Attributes:
        IEX (str): Investor's exchange data feed
        SIP (str): Securities Information Processor feed
        OTC (str): Over the counter feed
    """

    IEX = "iex"
    SIP = "sip"
    OTC = "otc"


class Adjustment(str, Enum):
    """Data normalization based on types of corporate actions.

    Attributes:
        RAW (str): Unadjusted data
        SPLIT (str): Stock-split adjusted data
        DIVIDEND (str): Dividend adjusted data
        ALL (str): Data adjusted for all corporate actions
    """

    RAW = "raw"
    SPLIT = "split"
    DIVIDEND = "dividend"
    ALL = "all"


class CryptoFeed(str, Enum):
    """
    Crypto location

    Attributes:
        US (str): United States crypto feed
    """

    US = "us"


class OptionsFeed(str, Enum):
    """
    The source feed of the data.
    `opra` requires subscription

    Attributes:
        OPRA (str): Options Price Reporting Authority
        INDICATIVE (str): Indicative data
    """

    OPRA = "opra"
    INDICATIVE = "indicative"


class MostActivesBy(str, Enum):
    """
    Most actives possible filters.

    Attributes:
        volume (str):
            Retrieve most actives by trading volume.
        trades (str):
            Retrieve most actives by number of trades.
    """

    VOLUME = "volume"
    TRADES = "trades"


class MarketType(str, Enum):
    """
    Most actives possible filters.

    Attributes:
        stocks (str)
        crypto (str)
    """

    STOCKS = "stocks"
    CRYPTO = "crypto"


class NewsImageSize(str, Enum):
    THUMB = "thumb"
    SMALL = "small"
    LARGE = "large"



from typing import Awaitable, Callable, Dict, Optional, Union

from alpaca.common.enums import BaseURL
from alpaca.data.enums import OptionsFeed
from alpaca.data.live.websocket import DataStream
from alpaca.data.models.quotes import Quote
from alpaca.data.models.trades import Trade


class OptionDataStream(DataStream):
    """
    A WebSocket client for streaming live option data.
    """

    def __init__(
        self,
        api_key: str,
        secret_key: str,
        raw_data: bool = False,
        feed: OptionsFeed = OptionsFeed.INDICATIVE,
        websocket_params: Optional[Dict] = None,
        url_override: Optional[str] = None,
    ) -> None:
        """
        Instantiates a WebSocket client for accessing live option data.

        Args:
            api_key (str): Alpaca API key.
            secret_key (str): Alpaca API secret key.
            raw_data (bool): Whether to return wrapped data or raw API data. Defaults to False.
            feed (OptionsFeed): The source feed of the data. `opra` or `indicative`.
                Defaults to `indicative`. `opra` requires a subscription.
            websocket_params (Optional[Dict], optional): Any parameters for configuring websocket
                connection. Defaults to None.
            url_override (Optional[str]): If specified allows you to override the base url the client
                points to for proxy/testing. Defaults to None.
        """
        super().__init__(
            endpoint=(
                url_override
                if url_override is not None
                else BaseURL.MARKET_DATA_STREAM.value + "/v1beta1/" + feed.value
            ),
            api_key=api_key,
            secret_key=secret_key,
            raw_data=raw_data,
            websocket_params=websocket_params,
        )

    def subscribe_trades(
        self, handler: Callable[[Union[Trade, Dict]], Awaitable[None]], *symbols: str
    ) -> None:
        """Subscribe to trades.

        Args:
            handler (Callable[[Union[Trade, Dict]], Awaitable[None]]): The coroutine callback
                function to handle the incoming data.
            *symbols: List of ticker symbols to subscribe to. "*" for everything.
        """
        self._subscribe(handler, symbols, self._handlers["trades"])

    def subscribe_quotes(
        self, handler: Callable[[Union[Quote, Dict]], Awaitable[None]], *symbols: str
    ) -> None:
        """Subscribe to quotes

        Args:
            handler (Callable[[Union[Quote, Dict]], Awaitable[None]]): The coroutine callback
                function to handle the incoming data.
            *symbols: List of ticker symbols to subscribe to. "*" for everything.
        """
        self._subscribe(handler, symbols, self._handlers["quotes"])

    def unsubscribe_trades(self, *symbols: str) -> None:
        """Unsubscribe from trades

        Args:
            *symbols (str): List of ticker symbols to unsubscribe from. "*" for everything.
        """
        self._unsubscribe("trades", symbols)

    def unsubscribe_quotes(self, *symbols: str) -> None:
        """Unsubscribe from quotes

        Args:
            *symbols (str): List of ticker symbols to unsubscribe from. "*" for everything.
        """
        self._unsubscribe("quotes", symbols)



import itertools
from typing import Any, Dict, List

import pandas as pd
from pandas import DataFrame

from alpaca.common.models import ValidateBaseModel as BaseModel


class TimeSeriesMixin:
    @property
    def df(self) -> DataFrame:
        """Returns a pandas dataframe containing the bar data.
        Requires mapping to be defined in child class.

        Returns:
            DataFrame: data in a pandas dataframe
        """
        # combine all lists of data into one list
        data_list = list(itertools.chain.from_iterable(self.dict().values()))

        df = pd.DataFrame(data_list)

        # set multi-level index
        if "news" in self.dict():
            # level=0 - id
            df = df.set_index(["id"])
        if set(["symbol", "timestamp"]).issubset(df.columns):
            # level=0 - symbol
            # level=1 - timestamp
            df = df.set_index(["symbol", "timestamp"])

        # drop null columns
        df.dropna(axis=1, how="all", inplace=True)

        return df


class BaseDataSet(BaseModel):
    """
    Base class to process data models for trades, bars quotes, and news.
    """

    data: Dict[str, List[BaseModel]] = {}

    def __getitem__(self, symbol: str) -> Any:
        """Gives dictionary-like access to multi-symbol data

        Args:
            symbol (str): The ticker identifier for the desired data

        Raises:
            KeyError: Symbol does not exist for data

        Returns:
            List[Bar]: The data for the given symbol
        """
        if symbol not in self.data:
            raise KeyError(f"No key {symbol} was found.")

        return self.data[symbol]

    def dict(self, **kwargs) -> dict:
        """
        Gives dictionary representation of data.

        Returns:
            dict: The data in dictionary form.
        """
        # converts each data (Bar, Quote, etc) in the symbol specific lists to its dictionary format
        return {
            symbol: list(map(lambda d: d.model_dump(), data_list))
            for symbol, data_list in self.data.items()
        }
